# stdlib
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import torch as th

# relative
from .....common.group import VerifyAll
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import SignedImmediateSyftMessageWithoutReply
from .....common.uid import UID
from .....store.storeable_object import StorableObject
from ....abstract.node import AbstractNode
from ...action.save_object_action import SaveObjectAction
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .tensor_manager_messages import CreateTensorMessage
from .tensor_manager_messages import CreateTensorResponse
from .tensor_manager_messages import DeleteTensorMessage
from .tensor_manager_messages import DeleteTensorResponse
from .tensor_manager_messages import GetTensorMessage
from .tensor_manager_messages import GetTensorResponse
from .tensor_manager_messages import GetTensorsMessage
from .tensor_manager_messages import GetTensorsResponse
from .tensor_manager_messages import UpdateTensorMessage
from .tensor_manager_messages import UpdateTensorResponse


def create_tensor_msg(
    msg: CreateTensorMessage,
    node: AbstractNode,
) -> CreateTensorResponse:
    try:
        payload = msg.content

        new_tensor = th.tensor(payload["tensor"])
        new_tensor.tag(*payload.get("tags", []))
        new_tensor.describe(payload.get("description", ""))

        id_at_location = UID()

        # Step 2: create message which contains object to send
        storable = StorableObject(
            id=id_at_location,
            data=new_tensor,
            tags=new_tensor.tags,
            description=new_tensor.description,
            search_permissions={VerifyAll(): None}
            if payload.get("searchable", False)
            else {},
        )

        obj_msg = SaveObjectAction(obj=storable, address=node.address)

        signed_message: SignedImmediateSyftMessageWithoutReply = obj_msg.sign(
            signing_key=SigningKey(
                payload["internal_key"].encode("utf-8"), encoder=HexEncoder
            )
        )

        node.recv_immediate_msg_without_reply(msg=signed_message)

        return CreateTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={
                "msg": "Tensor created succesfully!",
                "tensor_id": str(id_at_location.value),
            },
        )
    except Exception as e:
        return CreateTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"error": str(e)},
        )


def update_tensor_msg(
    msg: UpdateTensorMessage,
    node: AbstractNode,
) -> UpdateTensorResponse:
    try:
        payload = msg.content

        new_tensor = th.tensor(payload["tensor"])
        new_tensor.tag(*payload.get("tags", []))
        new_tensor.describe(payload.get("description", ""))

        key = UID.from_string(value=payload["tensor_id"])

        # Step 2: create message which contains object to send
        storable = StorableObject(
            id=key,
            data=new_tensor,
            tags=new_tensor.tags,
            description=new_tensor.description,
            search_permissions={VerifyAll(): None}
            if payload.get("searchable", False)
            else {},
        )

        obj_msg = SaveObjectAction(obj=storable, address=node.address)

        signed_message: SignedImmediateSyftMessageWithoutReply = obj_msg.sign(
            signing_key=SigningKey(
                payload["internal_key"].encode("utf-8"), encoder=HexEncoder
            )
        )

        node.recv_immediate_msg_without_reply(msg=signed_message)

        return UpdateTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"msg": "Tensor modified successfully!"},
        )
    except Exception as e:
        return UpdateTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"error": str(e)},
        )


def get_tensor_msg(
    msg: GetTensorMessage,
    node: AbstractNode,
) -> GetTensorResponse:
    try:
        payload = msg.content

        # Retrieve the dataset from node.store
        key = UID.from_string(value=payload["tensor_id"])
        tensor = node.store[key]
        return GetTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={
                "tensor": {
                    "id": payload["tensor_id"],
                    "tags": tensor.tags,
                    "description": tensor.description,
                }
            },
        )
    except Exception as e:
        return GetTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"error": str(e)},
        )


def get_tensors_msg(
    msg: GetTensorsMessage,
    node: AbstractNode,
) -> GetTensorsResponse:
    try:
        tensors = node.store.get_objects_of_type(obj_type=th.Tensor)

        result = []

        for tensor in tensors:
            result.append(
                {
                    "id": str(tensor.id.value),
                    "tags": tensor.tags,
                    "description": tensor.description,
                }
            )
        return GetTensorsResponse(
            address=msg.reply_to,
            status_code=200,
            content={"tensors": result},
        )
    except Exception as e:
        return GetTensorsResponse(
            address=msg.reply_to, content={"error": str(e)}, status_code=200
        )


def del_tensor_msg(
    msg: DeleteTensorMessage,
    node: AbstractNode,
) -> DeleteTensorResponse:
    try:
        payload = msg.content

        # Retrieve the dataset from node.store
        key = UID.from_string(value=payload["tensor_id"])
        node.store.delete(key=key)

        return DeleteTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"msg": "Tensor deleted successfully!"},
        )
    except Exception as e:
        return DeleteTensorResponse(
            address=msg.reply_to,
            status_code=200,
            content={"error": str(e)},
        )


class TensorManagerService(ImmediateNodeServiceWithReply):
    INPUT_TYPE = Union[
        Type[CreateTensorMessage],
        Type[UpdateTensorMessage],
        Type[GetTensorMessage],
        Type[GetTensorsMessage],
        Type[DeleteTensorMessage],
    ]

    INPUT_MESSAGES = Union[
        CreateTensorMessage,
        UpdateTensorMessage,
        GetTensorMessage,
        GetTensorsMessage,
        DeleteTensorMessage,
    ]

    OUTPUT_MESSAGES = Union[
        CreateTensorResponse,
        UpdateTensorResponse,
        GetTensorResponse,
        GetTensorsResponse,
        DeleteTensorResponse,
    ]

    msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateTensorMessage: create_tensor_msg,
        UpdateTensorMessage: update_tensor_msg,
        GetTensorMessage: get_tensor_msg,
        GetTensorsMessage: get_tensors_msg,
        DeleteTensorMessage: del_tensor_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        return TensorManagerService.msg_handler_map[type(msg)](msg=msg, node=node)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateTensorMessage,
            UpdateTensorMessage,
            GetTensorMessage,
            GetTensorsMessage,
            DeleteTensorMessage,
        ]



# future
from __future__ import annotations

# stdlib
from collections.abc import Sequence
import os
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import capnp
from nacl.signing import VerifyKey
import numpy as np

# relative
from ....core.adp.entity import Entity
from ....core.adp.entity_list import EntityList
from ....lib.numpy.array import arrow_deserialize as numpy_deserialize
from ....lib.numpy.array import arrow_serialize as numpy_serialize
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.deserialize import CAPNP_END_MAGIC_HEADER_BYTES
from ...common.serde.deserialize import CAPNP_START_MAGIC_HEADER_BYTES
from ...common.serde.serializable import serializable
from ...common.uid import UID
from ...pointer.pointer import Pointer
from ..ancestors import AutogradTensorAncestor
from ..lazy_repeat_array import lazyrepeatarray
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import SupportedChainType  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore
from .adp_tensor import ADPTensor
from .initial_gamma import InitialGammaTensor
from .initial_gamma import IntermediateGammaTensor
from .single_entity_phi import SingleEntityPhiTensor


@serializable(recursive_serde=True)
class TensorWrappedNDimEntityPhiTensorPointer(Pointer):
    __name__ = "TensorWrappedNDimEntityPhiTensorPointer"
    __module__ = "syft.core.tensor.autodp.ndim_entity_phi"
    __attr_allowlist__ = ("min_vals", "max_vals", "entities")
    # TODO :should create serialization for Entity List

    def __init__(
        self,
        entities: EntityList,
        min_vals: np.typing.ArrayLike,
        max_vals: np.typing.ArrayLike,
        client: Any,
        # scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[np.dtype] = None,
    ):
        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.min_vals = min_vals
        self.max_vals = max_vals
        self.entities = entities
        # self.scalar_manager = scalar_manager
        self.public_shape = public_shape
        self.public_dtype = public_dtype


@serializable(capnp_bytes=True)
class NDimEntityPhiTensor(PassthroughTensor, AutogradTensorAncestor, ADPTensor):
    PointerClassOverride = TensorWrappedNDimEntityPhiTensorPointer
    __attr_allowlist__ = ["child", "min_vals", "max_vals", "entities"]
    __slots__ = (
        "child",
        "min_vals",
        "max_vals",
        "entities",
    )

    def __init__(
        self,
        child: Sequence,
        entities: Union[List[Entity], EntityList],
        min_vals: np.ndarray,
        max_vals: np.ndarray,
        row_type: SingleEntityPhiTensor = SingleEntityPhiTensor,  # type: ignore
    ) -> None:

        # child = the actual private data
        super().__init__(child)

        # lazyrepeatarray matching the shape of child
        if not isinstance(min_vals, lazyrepeatarray):
            min_vals = lazyrepeatarray(data=min_vals, shape=child.shape)  # type: ignore
        if not isinstance(max_vals, lazyrepeatarray):
            max_vals = lazyrepeatarray(data=max_vals, shape=child.shape)  # type: ignore
        self.min_vals = min_vals
        self.max_vals = max_vals

        if not isinstance(entities, EntityList):
            entities = EntityList.from_objs(entities)

        self.entities = entities

        # if scalar_manager is None:
        #     self.scalar_manager = VirtualMachinePrivateScalarManager()
        # else:
        #     self.scalar_manager = scalar_manager

    @staticmethod
    def from_rows(rows: Sequence) -> NDimEntityPhiTensor:
        if len(rows) < 1 or not isinstance(rows[0], SingleEntityPhiTensor):
            raise Exception(
                "NDimEntityPhiTensor.from_rows requires a list of SingleEntityPhiTensors"
            )

        # create lazyrepeatarrays of the first element
        first_row = rows[0]
        min_vals = lazyrepeatarray(
            data=first_row.min_vals,
            shape=tuple([len(rows)] + list(first_row.min_vals.shape)),
        )
        max_vals = lazyrepeatarray(
            data=first_row.max_vals,
            shape=tuple([len(rows)] + list(first_row.max_vals.shape)),
        )

        # collect entities and children into numpy arrays
        entity_list = []
        child_list = []
        for row in rows:
            entity_list.append(row.entity)
            child_list.append(row.child)
        entities = EntityList.from_objs(entities=entity_list)
        child = np.stack(child_list)

        # use new constructor
        return NDimEntityPhiTensor(
            child=child,
            min_vals=min_vals,
            max_vals=max_vals,
            entities=entities,
        )

    def init_pointer(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> TensorWrappedNDimEntityPhiTensorPointer:
        return TensorWrappedNDimEntityPhiTensorPointer(
            # Arguments specifically for SEPhiTensor
            entities=self.entities,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            # scalar_manager=self.scalar_manager,
            # Arguments required for a Pointer to work
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

    @property
    def gamma(self) -> InitialGammaTensor:
        """Property to cast this tensor into a GammaTensor"""
        return self.create_gamma()

    def copy(self, order: Optional[str] = "K") -> NDimEntityPhiTensor:
        """Return copy of the given object"""

        return NDimEntityPhiTensor(
            child=self.child.copy(order=order),
            min_vals=self.min_vals.copy(order=order),
            max_vals=self.max_vals.copy(order=order),
            entities=self.entities.copy(order=order),
        )

    def all(self) -> bool:
        return self.child.all()

    def any(self) -> bool:
        return self.child.any()

    def copy_with(self, child: np.ndarray) -> NDimEntityPhiTensor:
        new_tensor = self.copy()
        new_tensor.child = child
        return new_tensor

    def create_gamma(
        self, scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None
    ) -> InitialGammaTensor:
        """Return a new Gamma tensor based on this phi tensor"""

        # if scalar_manager is None:
        #     scalar_manager = self.scalar_manager

        # Gamma expects an entity for each scalar
        # entities = np.array([self.entity] * np.array(self.child.shape).prod()).reshape(
        #     self.shape
        # )

        # TODO: update InitialGammaTensor to handle EntityList
        return InitialGammaTensor(
            values=self.child,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            entities=self.entities,
            # scalar_manager=scalar_manager,
        )

    def publish(
        self, acc: Any, sigma: float, user_key: VerifyKey
    ) -> AcceptableSimpleType:
        print("PUBLISHING TO GAMMA:")
        print(self.child)
        return self.gamma.publish(acc=acc, sigma=sigma, user_key=user_key)

    @property
    def value(self) -> np.ndarray:
        return self.child

    def astype(self, np_type: np.dtype) -> NDimEntityPhiTensor:
        return self.__class__(
            child=self.child.astype(np_type),
            entities=self.entities,
            min_vals=self.min_vals.astype(np_type),
            max_vals=self.max_vals.astype(np_type),
            # scalar_manager=self.scalar_manager,
        )

    @property
    def shape(self) -> Tuple[Any, ...]:
        return self.child.shape

    def __repr__(self) -> str:
        """Pretty print some information, optimized for Jupyter notebook viewing."""
        return (
            f"{self.__class__.__name__}(child={self.child.shape}, "
            + f"min_vals={self.min_vals}, max_vals={self.max_vals})"
        )

    def __eq__(self, other: Any) -> Union[NDimEntityPhiTensor, IntermediateGammaTensor]:
        # TODO: what about entities and min / max values?
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            gamma_output = False
            if is_acceptable_simple_type(other):
                result = self.child == other
            else:
                # check entities match, if they dont gamma_output = True
                #
                result = self.child == other.child
                if isinstance(result, InitialGammaTensor):
                    gamma_output = True
            if not gamma_output:
                # min_vals=self.min_vals * 0.0,
                # max_vals=self.max_vals * 0.0 + 1.0,
                return self.copy_with(child=result)
            else:
                return self.copy_with(child=result).gamma
        else:
            raise Exception(
                "Tensor dims do not match for __eq__: "
                + f"{len(self.child)} != {len(other.child)}"
            )

    def __add__(
        self, other: SupportedChainType
    ) -> Union[NDimEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being added is also private
        if isinstance(other, NDimEntityPhiTensor):
            if self.entities != other.entities:
                return self.gamma + other.gamma

            return NDimEntityPhiTensor(
                child=self.child + other.child,
                min_vals=self.min_vals + other.min_vals,
                max_vals=self.max_vals + other.max_vals,
                entities=self.entities,
                # scalar_manager=self.scalar_manager,
            )

        # if the tensor being added is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):
            return NDimEntityPhiTensor(
                child=self.child + other,
                min_vals=self.min_vals + other,
                max_vals=self.max_vals + other,
                entities=self.entities,
                # scalar_manager=self.scalar_manager,
            )

        elif isinstance(other, IntermediateGammaTensor):
            return self.gamma + other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    @staticmethod
    def get_capnp_schema() -> type:
        here = os.path.dirname(__file__)
        root_dir = Path(here) / ".." / ".." / ".." / ".." / ".." / "capnp"
        return capnp.load(str(root_dir / "ndept.capnp"))

    @staticmethod
    def chunk_bytes(
        data: Sequence, field_name: str, builder: capnp.lib.capnp._DynamicStructBuilder
    ) -> List:
        CHUNK_SIZE = int(5.12e8)  # capnp max for a List(Data) field
        list_size = len(data) // CHUNK_SIZE + 1
        data_lst = builder.init(field_name, list_size)
        END_INDEX = CHUNK_SIZE
        for idx in range(list_size):
            START_INDEX = idx * CHUNK_SIZE
            END_INDEX = min(START_INDEX + CHUNK_SIZE, len(data))
            data_lst[idx] = data[START_INDEX:END_INDEX]

        return data_lst

    @staticmethod
    def combine_bytes(capnp_list: List[bytes]) -> bytes:
        # TODO: make sure this doesn't copy, perhaps allocate a fixed size buffer
        # and move the bytes into it as we go
        bytes_value = b""
        for value in capnp_list:
            bytes_value += value
        return bytes_value

    @staticmethod
    def serde_magic_header() -> bytes:
        return (
            CAPNP_START_MAGIC_HEADER_BYTES
            + NDimEntityPhiTensor.__name__.encode("utf-8")
            + CAPNP_END_MAGIC_HEADER_BYTES
        )

    def _object2bytes(self) -> bytes:
        schema = NDimEntityPhiTensor.get_capnp_schema()

        rows, rows_size = numpy_serialize(self.child, get_bytes=True)
        min_vals, min_vals_size = numpy_serialize(self.min_vals.data, get_bytes=True)
        max_vals, max_vals_size = numpy_serialize(self.max_vals.data, get_bytes=True)
        entities_indexed, entities_indexed_size = numpy_serialize(
            self.entities.entities_indexed, get_bytes=True
        )
        one_hot_lookup = self.entities.one_hot_lookup

        ndept_struct: capnp.lib.capnp._StructModule = schema.NDEPT  # type: ignore
        ndept_msg = ndept_struct.new_message()
        metadata_schema = ndept_struct.TensorMetadata
        child_metadata = metadata_schema.new_message()
        min_vals_metadata = metadata_schema.new_message()
        max_vals_metadata = metadata_schema.new_message()
        entities_metadata = metadata_schema.new_message()

        # this is how we dispatch correct deserialization of bytes
        ndept_msg.magicHeader = NDimEntityPhiTensor.serde_magic_header()

        NDimEntityPhiTensor.chunk_bytes(rows, "child", ndept_msg)
        child_metadata.dtype = str(self.child.dtype)
        child_metadata.decompressedSize = rows_size
        ndept_msg.childMetadata = child_metadata

        NDimEntityPhiTensor.chunk_bytes(min_vals, "minVals", ndept_msg)
        min_vals_metadata.dtype = str(self.min_vals.data.dtype)
        min_vals_metadata.decompressedSize = min_vals_size
        ndept_msg.minValsMetadata = min_vals_metadata

        NDimEntityPhiTensor.chunk_bytes(max_vals, "maxVals", ndept_msg)
        max_vals_metadata.dtype = str(self.max_vals.data.dtype)
        max_vals_metadata.decompressedSize = max_vals_size
        ndept_msg.maxValsMetadata = max_vals_metadata

        NDimEntityPhiTensor.chunk_bytes(entities_indexed, "entitiesIndexed", ndept_msg)
        entities_metadata.dtype = str(self.entities.entities_indexed.dtype)
        entities_metadata.decompressedSize = entities_indexed_size
        ndept_msg.entitiesIndexedMetadata = entities_metadata

        oneHotLookupList = ndept_msg.init("oneHotLookup", len(one_hot_lookup))
        for i, entity in enumerate(one_hot_lookup):
            oneHotLookupList[i] = (
                entity if not getattr(entity, "name", None) else entity.name  # type: ignore
            )

        # to pack or not to pack?
        # return ndept_msg.to_bytes()
        return ndept_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> NDimEntityPhiTensor:
        schema = NDimEntityPhiTensor.get_capnp_schema()
        ndept_struct: capnp.lib.capnp._StructModule = schema.NDEPT  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        # ndept_msg = ndept_struct.from_bytes(buf, traversal_limit_in_words=2 ** 64 - 1)
        ndept_msg = ndept_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        child_metadata = ndept_msg.childMetadata

        child = numpy_deserialize(
            NDimEntityPhiTensor.combine_bytes(ndept_msg.child),
            child_metadata.decompressedSize,
            child_metadata.dtype,
        )

        min_vals_metadata = ndept_msg.minValsMetadata
        min_vals = lazyrepeatarray(
            numpy_deserialize(
                NDimEntityPhiTensor.combine_bytes(ndept_msg.minVals),
                min_vals_metadata.decompressedSize,
                min_vals_metadata.dtype,
            ),
            child.shape,
        )

        max_vals_metadata = ndept_msg.maxValsMetadata
        max_vals = lazyrepeatarray(
            numpy_deserialize(
                NDimEntityPhiTensor.combine_bytes(ndept_msg.maxVals),
                max_vals_metadata.decompressedSize,
                max_vals_metadata.dtype,
            ),
            child.shape,
        )

        entities_metadata = ndept_msg.entitiesIndexedMetadata
        entities_indexed = numpy_deserialize(
            NDimEntityPhiTensor.combine_bytes(ndept_msg.entitiesIndexed),
            entities_metadata.decompressedSize,
            entities_metadata.dtype,
        )
        one_hot_lookup = np.array(ndept_msg.oneHotLookup)

        entity_list = EntityList(one_hot_lookup, entities_indexed)

        return NDimEntityPhiTensor(
            child=child, min_vals=min_vals, max_vals=max_vals, entities=entity_list
        )



# stdlib
from copy import deepcopy
from typing import Any
from typing import Collection
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union
from typing import cast

# third party
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

# syft absolute
import syft as sy

# relative
from ....common.uid import UID
from ....node.common.node_table.bin_obj_dataset import BinObjDataset
from ....store import ObjectStore
from ....store.storeable_object import StorableObject
from ..node_table.bin_obj_metadata import ObjectMetadata


class DictStore(ObjectStore):
    def __init__(self, db: Session) -> None:
        self.db = db
        self.kv_store: dict[UID, Any] = {}

    def get_object(self, key: UID) -> Optional[StorableObject]:
        try:
            return self.__getitem__(key)
        except KeyError as e:  # noqa: F841
            return None

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        # raise NotImplementedError("get_objects_of_type")
        # return [obj for obj in self.values() if isinstance(obj.data, obj_type)]
        return self.values()

    def __sizeof__(self) -> int:
        return self.values().__sizeof__()

    def __str__(self) -> str:
        return f"{type(self)}"

    def __len__(self) -> int:
        return len(self.kv_store.keys())

    def keys(self) -> Collection[UID]:
        return self.kv_store.keys()

    def values(self) -> List[StorableObject]:
        keys = self.kv_store.keys()
        # this is bad we need to decouple getting the data from the search
        all_values = []
        for key in keys:
            all_values.append(self.__getitem__(key))

        return all_values

    def __contains__(self, key: UID) -> bool:
        return key in self.kv_store

    def __getitem__(self, key: Union[UID, str, bytes]) -> StorableObject:
        try:
            local_session = sessionmaker(bind=self.db)()

            if isinstance(key, UID):
                key_str = str(key.value)
                key_uid = key
            elif isinstance(key, bytes):
                key_str = str(key.decode("utf-8"))
                key_uid = UID.from_string(key_str)
            else:
                key_str = key
                key_uid = UID.from_string(key_str)

            store_obj = self.kv_store[key_uid]

            # serialized contents
            if isinstance(store_obj, bytes):
                try:
                    de = sy.deserialize(store_obj, from_bytes=True)
                    obj = de
                except Exception as e:
                    raise Exception(f"Failed to deserialize obj at key {key_str}. {e}")
            else:
                # not serialized
                try:
                    obj = deepcopy(store_obj)
                except Exception as e:
                    raise Exception(
                        f"DictStore should not contain unpickleable objects. {e}"
                    )

            if id(obj) == id(store_obj):
                raise Exception("Objects must use deepcopy or mutation can occur")

            obj_metadata = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).first()
            )

            if obj is None or obj_metadata is None:
                raise KeyError(f"Object not found! for UID: {key_uid}")

            obj = StorableObject(
                id=key_uid,
                data=obj.data,
                description=obj_metadata.description,
                tags=obj_metadata.tags,
                read_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.read_permissions), from_bytes=True
                    )
                ),
                search_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.search_permissions), from_bytes=True
                    )
                ),
                write_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.write_permissions), from_bytes=True
                    )
                ),
            )
            local_session.close()
            return obj
        except Exception as e:
            print(f"Cant get object {str(key)}", e)
            raise KeyError(f"Object not found! for UID: {str(key)}")

    def is_dataset(self, key: UID) -> bool:
        local_session = sessionmaker(bind=self.db)()
        is_dataset_obj = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).exists()
        )
        is_dataset_obj = local_session.query(is_dataset_obj).scalar()
        local_session.close()
        return is_dataset_obj

    def _get_obj_dataset_relation(self, key: UID) -> Optional[BinObjDataset]:
        local_session = sessionmaker(bind=self.db)()
        obj_dataset_relation = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).first()
        )
        local_session.close()
        return obj_dataset_relation

    def __setitem__(self, key: UID, value: StorableObject) -> None:
        key_str = str(key.value)
        try:
            store_obj = deepcopy(value)
            self.kv_store[key] = store_obj
            if id(value) == id(store_obj):
                raise Exception("Objects must use deepcopy or mutation can occur")
        except Exception as e:
            # if we get a pickling error from deepcopy we can juse use sy.serialize
            # in theory not having to do this unless necessary means it should be a
            # little faster and less memory intensive
            if "Pickling" in str(e):
                try:
                    # deepcopy falls back to pickle and some objects can't be pickled
                    self.kv_store[key] = sy.serialize(value, to_bytes=True)
                except Exception as ex:
                    raise Exception(f"Failed to serialize object {type(value)}. {ex}")
            else:
                raise Exception(f"Failed to save object {type(value)}. {e}")

        create_metadata = True
        local_session = sessionmaker(bind=self.db)()
        try:
            # use existing metadata row to prevent more than 1
            metadata_obj = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).all()[-1]
            )
            create_metadata = False
        except Exception:
            pass
            # no metadata row exists lets insert one
            metadata_obj = ObjectMetadata()

        metadata_obj.obj = str(key.value)
        metadata_obj.tags = value.tags
        metadata_obj.description = value.description
        metadata_obj.read_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.read_permissions), to_bytes=True),
        ).hex()
        metadata_obj.search_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.search_permissions), to_bytes=True),
        ).hex()
        metadata_obj.write_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.write_permissions), to_bytes=True),
        ).hex()

        obj_dataset_relation = self._get_obj_dataset_relation(key)
        if obj_dataset_relation:
            # Create a object dataset relationship for the new object
            obj_dataset_relation = BinObjDataset(
                # id=obj_dataset_relation.id,  NOTE: Commented temporarily
                name=obj_dataset_relation.name,
                obj=str(key.value),
                dataset=obj_dataset_relation.dataset,
                dtype=obj_dataset_relation.dtype,
                shape=obj_dataset_relation.shape,
            )

        if create_metadata:
            local_session.add(metadata_obj)
        local_session.add(obj_dataset_relation) if obj_dataset_relation else None
        local_session.commit()
        local_session.close()

    def delete(self, key: UID) -> None:
        try:
            del self.kv_store[key]
            local_session = sessionmaker(bind=self.db)()
            metadata_to_delete = (
                local_session.query(ObjectMetadata)
                .filter_by(obj=str(key.value))
                .first()
            )
            local_session.delete(metadata_to_delete)
            local_session.commit()
            local_session.close()
        except Exception as e:
            print(f"{type(self)} Exception in __delitem__ error {key}. {e}")

    def clear(self) -> None:
        self.kv_store = {}
        local_session = sessionmaker(bind=self.db)()
        local_session.query(ObjectMetadata).delete()
        local_session.commit()
        local_session.close()

    def __repr__(self) -> str:
        return f"{type(self)}"



# stdlib
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey

# relative
from ......proto.core.io.address_pb2 import Address as Address_PB
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.uid import UID
from ....domain.domain import Domain
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..node_service import ImmediateNodeServiceWithoutReply
from .object_transfer_messages import LoadObjectMessage
from .object_transfer_messages import LoadObjectResponse
from .object_transfer_messages import SaveObjectMessage
from .object_transfer_messages import SaveObjectResponse


class TransferObjectService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: Domain,
        msg: Union[
            LoadObjectMessage,
        ],
        verify_key: VerifyKey,
    ) -> Union[LoadObjectResponse, SaveObjectResponse]:
        _worker_address = msg.content.get("address", None)
        _obj_id = msg.content.get("uid", None)
        _current_user_id = msg.content.get("current_user", None)

        users = node.users

        if not _current_user_id:
            _current_user_id = users.first(
                verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
            ).id

        addr_pb = Address_PB()
        addr_pb.ParseFromString(_worker_address.encode("ISO-8859-1"))
        _syft_address = _deserialize(blob=addr_pb)

        _syft_id = UID.from_string(value=_obj_id)

        _worker_client = node.in_memory_client_registry[_syft_address.domain_id]

        try:
            _obj = node.store[_syft_id]
        except Exception:
            raise Exception("Object Not Found!")

        _obj.data.send(
            _worker_client,
            searchable=True,
            tags=_obj.tags,
            description=_obj.description,
        )

        return LoadObjectResponse(
            address=msg.reply_to,
            status_code=200,
            content={"msg": "Object loaded successfully!"},
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [LoadObjectMessage]


class SaveObjectService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: Domain,
        msg: SaveObjectMessage,
        verify_key: VerifyKey,
    ) -> None:
        _obj_id = msg.content.get("uid", None)

        _syft_id = UID.from_string(value=_obj_id)

        try:
            _obj = node.store[_syft_id]  # noqa: 841
        except Exception:
            raise Exception("Object Not Found!")

        # TODO: this should call something using the 'node' variable. not executor directly (which is a pygrid op)
        # executor.submit(send_obj, _obj, node)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithoutReply]]:
        return [SaveObjectMessage]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/metadata.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%proto/core/node/common/metadata.proto\x12\x15syft.core.node.common\x1a%proto/core/common/common_object.proto\x1a\x1cproto/core/io/location.proto"\x8d\x01\n\x08Metadata\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0c\n\x04name\x18\x02 \x01(\t\x12,\n\x04node\x18\x03 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x11\n\tnode_type\x18\x04 \x01(\t\x12\x0f\n\x07version\x18\x05 \x01(\tb\x06proto3'
)


_METADATA = DESCRIPTOR.message_types_by_name["Metadata"]
Metadata = _reflection.GeneratedProtocolMessageType(
    "Metadata",
    (_message.Message,),
    {
        "DESCRIPTOR": _METADATA,
        "__module__": "proto.core.node.common.metadata_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.Metadata)
    },
)
_sym_db.RegisterMessage(Metadata)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _METADATA._serialized_start = 134
    _METADATA._serialized_end = 275
# @@protoc_insertion_point(module_scope)



# future
from __future__ import annotations

# stdlib
import copy
from typing import Optional
from typing import Union
from urllib.parse import urlparse

# third party
import requests

# relative
from ..core.common.serde.serializable import serializable
from ..util import verify_tls


@serializable(recursive_serde=True)
class GridURL:
    __attr_allowlist__ = ["protocol", "host_or_ip", "port", "path"]

    @staticmethod
    def from_url(url: Union[str, GridURL]) -> GridURL:
        if isinstance(url, GridURL):
            return url
        try:
            # urlparse doesnt handle no protocol properly
            if "://" not in url:
                url = "http://" + url
            parts = urlparse(url)
            host_or_ip_parts = parts.netloc.split(":")
            # netloc is host:port
            port = 80
            if len(host_or_ip_parts) > 1:
                port = int(host_or_ip_parts[1])
            host_or_ip = host_or_ip_parts[0]
            return GridURL(
                host_or_ip=host_or_ip, path=parts.path, port=port, protocol=parts.scheme
            )
        except Exception as e:
            print(f"Failed to convert url: {url} to GridURL. {e}")
            raise e

    def __init__(
        self,
        protocol: str = "http",
        host_or_ip: str = "localhost",
        port: Optional[int] = 80,
        path: str = "",
    ) -> None:
        # in case a preferred port is listed but its not clear if an alternative
        # port was included in the supplied host_or_ip:port combo passed in earlier
        if ":" in host_or_ip:
            sub_grid_url: GridURL = GridURL.from_url(host_or_ip)
            host_or_ip = str(sub_grid_url.host_or_ip)  # type: ignore
            port = int(sub_grid_url.port)  # type: ignore
            protocol = str(sub_grid_url.protocol)  # type: ignore
            path = str(sub_grid_url.path)  # type: ignore
        elif port is None:
            port = 80

        self.host_or_ip = host_or_ip
        self.path = path
        self.port = port
        self.protocol = protocol

    def with_path(self, path: str) -> GridURL:
        dupe = copy.copy(self)
        dupe.path = path
        return dupe

    def as_docker_host(self) -> GridURL:
        if self.host_or_ip != "localhost":
            return self
        return GridURL(
            protocol=self.protocol,
            host_or_ip="docker-host",
            port=self.port,
            path=self.path,
        )

    @property
    def url(self) -> str:
        return f"{self.base_url}{self.path}"

    @property
    def base_url(self) -> str:
        return f"{self.protocol}://{self.host_or_ip}:{self.port}"

    def to_tls(self) -> GridURL:
        if self.protocol == "https":
            return self

        # TODO: only ignore ssl in dev mode
        r = requests.get(
            self.base_url, verify=verify_tls()
        )  # ignore ssl cert if its fake
        new_base_url = r.url
        if new_base_url.endswith("/"):
            new_base_url = new_base_url[0:-1]
        return GridURL.from_url(url=f"{new_base_url}{self.path}")

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.url}>"

    def __str__(self) -> str:
        return self.url

    def __hash__(self) -> int:
        return hash(self.__str__())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/util/data_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1dproto/util/data_message.proto\x12\tsyft.util"0\n\x0b\x44\x61taMessage\x12\x10\n\x08obj_type\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\x62\x06proto3'
)


_DATAMESSAGE = DESCRIPTOR.message_types_by_name["DataMessage"]
DataMessage = _reflection.GeneratedProtocolMessageType(
    "DataMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATAMESSAGE,
        "__module__": "proto.util.data_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.util.DataMessage)
    },
)
_sym_db.RegisterMessage(DataMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DATAMESSAGE._serialized_start = 44
    _DATAMESSAGE._serialized_end = 92
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import List as TypeList

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ......proto.core.node.domain.service.pss_pb2 import PSA  # type: ignore
from .....common.message import ImmediateSyftMessageWithoutReply  # type: ignore
from .....common.serde.deserialize import _deserialize as deserialize  # type: ignore
from .....common.serde.serializable import serializable  # type: ignore
from .....common.serde.serialize import _serialize as serialize  # type: ignore
from .....common.uid import UID  # type: ignore
from .....io.address import Address  # type: ignore


@serializable()
@final
class PublishScalarsAction(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        id_at_location: UID,
        address: Address,
        publish_ids_at_location: TypeList[UID],
        sigma: float,
    ):
        super().__init__(address=address)
        self.id_at_location = id_at_location
        self.publish_ids_at_location = publish_ids_at_location
        self.sigma = sigma

    def _object2proto(self) -> PSA:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: PublishScalarsAction_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return PSA(
            id_at_location=serialize(self.id_at_location),
            address=serialize(self.address),
            publish_ids_at_location=[
                serialize(uid) for uid in self.publish_ids_at_location
            ],
            sigma=self.sigma,
        )

    @staticmethod
    def _proto2object(proto: PSA) -> "PublishScalarsAction":
        """Creates a PublishScalarsAction from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of PublishScalarsAction
        :rtype: PublishScalarsAction

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return PublishScalarsAction(
            id_at_location=deserialize(blob=proto.id_at_location),
            address=deserialize(blob=proto.address),
            publish_ids_at_location=[
                deserialize(blob=ids) for ids in proto.publish_ids_at_location
            ],
            sigma=proto.sigma,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return PSA



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/network_search_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n1proto/grid/messages/network_search_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9f\x01\n\x14NetworkSearchMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x15NetworkSearchResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_NETWORKSEARCHMESSAGE = DESCRIPTOR.message_types_by_name["NetworkSearchMessage"]
_NETWORKSEARCHRESPONSE = DESCRIPTOR.message_types_by_name["NetworkSearchResponse"]
NetworkSearchMessage = _reflection.GeneratedProtocolMessageType(
    "NetworkSearchMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _NETWORKSEARCHMESSAGE,
        "__module__": "proto.grid.messages.network_search_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.NetworkSearchMessage)
    },
)
_sym_db.RegisterMessage(NetworkSearchMessage)

NetworkSearchResponse = _reflection.GeneratedProtocolMessageType(
    "NetworkSearchResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _NETWORKSEARCHRESPONSE,
        "__module__": "proto.grid.messages.network_search_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.NetworkSearchResponse)
    },
)
_sym_db.RegisterMessage(NetworkSearchResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _NETWORKSEARCHMESSAGE._serialized_start = 142
    _NETWORKSEARCHMESSAGE._serialized_end = 301
    _NETWORKSEARCHRESPONSE._serialized_start = 304
    _NETWORKSEARCHRESPONSE._serialized_end = 444
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/domain/service/request_handler_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n<proto/core/node/domain/service/request_handler_message.proto\x12\x1dsyft.core.node.domain.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto"\xa2\x01\n\x1bUpdateRequestHandlerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12&\n\x07handler\x18\x03 \x01(\x0b\x32\x15.syft.lib.python.Dict\x12\x0c\n\x04keep\x18\x04 \x01(\x08"\x96\x01\n\x1cGetAllRequestHandlersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n$GetAllRequestHandlersResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08handlers\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dictb\x06proto3'
)


_UPDATEREQUESTHANDLERMESSAGE = DESCRIPTOR.message_types_by_name[
    "UpdateRequestHandlerMessage"
]
_GETALLREQUESTHANDLERSMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAllRequestHandlersMessage"
]
_GETALLREQUESTHANDLERSRESPONSEMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAllRequestHandlersResponseMessage"
]
UpdateRequestHandlerMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateRequestHandlerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEREQUESTHANDLERMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.UpdateRequestHandlerMessage)
    },
)
_sym_db.RegisterMessage(UpdateRequestHandlerMessage)

GetAllRequestHandlersMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestHandlersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTHANDLERSMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestHandlersMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestHandlersMessage)

GetAllRequestHandlersResponseMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestHandlersResponseMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTHANDLERSRESPONSEMESSAGE,
        "__module__": "proto.core.node.domain.service.request_handler_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestHandlersResponseMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestHandlersResponseMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _UPDATEREQUESTHANDLERMESSAGE._serialized_start = 193
    _UPDATEREQUESTHANDLERMESSAGE._serialized_end = 355
    _GETALLREQUESTHANDLERSMESSAGE._serialized_start = 358
    _GETALLREQUESTHANDLERSMESSAGE._serialized_end = 508
    _GETALLREQUESTHANDLERSRESPONSEMESSAGE._serialized_start = 511
    _GETALLREQUESTHANDLERSRESPONSEMESSAGE._serialized_end = 669
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Collection
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union
from typing import cast

# third party
import redis
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

# syft absolute
import syft

# relative
from ....common.uid import UID
from ....node.common.node_table.bin_obj_dataset import BinObjDataset
from ....store import ObjectStore
from ....store.storeable_object import StorableObject
from ..node_table.bin_obj_metadata import ObjectMetadata


class RedisStore(ObjectStore):
    def __init__(self, db: Session) -> None:
        self.db = db
        try:
            # TODO: refactor hard coded host and port to configuration
            self.redis: redis.client.Redis = redis.Redis(host="redis", port=6379)
        except Exception as e:
            print("failed to load redis", e)
            raise e

    def get_object(self, key: UID) -> Optional[StorableObject]:
        try:
            return self.__getitem__(key)
        except KeyError as e:  # noqa: F841
            return None

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        # raise NotImplementedError("get_objects_of_type")
        # return [obj for obj in self.values() if isinstance(obj.data, obj_type)]
        return self.values()

    def __sizeof__(self) -> int:
        return self.values().__sizeof__()

    def __str__(self) -> str:
        return f"{type(self)}"

    def __len__(self) -> int:
        return self.redis.dbsize()

    def keys(self) -> Collection[UID]:
        key_bytes = self.redis.keys()
        key_ids = [UID.from_string(str(key.decode("utf-8"))) for key in key_bytes]
        return key_ids

    def values(self) -> List[StorableObject]:
        key_bytes = self.redis.keys()
        # this is bad we need to decouple getting the data from the search
        all_values = []
        for key in key_bytes:
            all_values.append(self.__getitem__(key))

        return all_values

    def __contains__(self, key: UID) -> bool:
        return str(key.value) in self.redis

    def __getitem__(self, key: Union[UID, str, bytes]) -> StorableObject:
        local_session = sessionmaker(bind=self.db)()

        if isinstance(key, UID):
            key_str = str(key.value)
            key_uid = key
        elif isinstance(key, bytes):
            key_str = str(key.decode("utf-8"))
            key_uid = UID.from_string(key_str)
        else:
            key_str = key
            key_uid = UID.from_string(key_str)

        obj = self.redis.get(key_str)
        obj_metadata = (
            local_session.query(ObjectMetadata).filter_by(obj=key_str).first()
        )
        if obj is None or obj_metadata is None:
            raise KeyError(f"Object not found! for UID: {key_str}")

        obj = syft.deserialize(obj, from_bytes=True)

        obj = StorableObject(
            id=key_uid,
            data=obj,
            description=obj_metadata.description,
            tags=obj_metadata.tags,
            read_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.read_permissions), from_bytes=True
                )
            ),
            search_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.search_permissions), from_bytes=True
                )
            ),
            write_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.write_permissions), from_bytes=True
                )
            ),
            # name=obj_metadata.name,
        )
        local_session.close()
        return obj

    def is_dataset(self, key: UID) -> bool:
        local_session = sessionmaker(bind=self.db)()
        is_dataset_obj = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).exists()
        )
        is_dataset_obj = local_session.query(is_dataset_obj).scalar()
        local_session.close()
        return is_dataset_obj

    def _get_obj_dataset_relation(self, key: UID) -> Optional[BinObjDataset]:
        local_session = sessionmaker(bind=self.db)()
        obj_dataset_relation = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).first()
        )
        local_session.close()
        return obj_dataset_relation

    def __setitem__(self, key: UID, value: StorableObject) -> None:
        bin = syft.serialize(value.data, to_bytes=True)
        self.redis.set(str(key.value), bin)

        key_str = str(key.value)
        create_metadata = True
        local_session = sessionmaker(bind=self.db)()
        try:
            # use existing metadata row to prevent more than 1
            metadata_obj = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).all()[-1]
            )
            create_metadata = False
        except Exception:
            pass
            # no metadata row exists lets insert one
            metadata_obj = ObjectMetadata()

        metadata_obj.obj = str(key.value)
        metadata_obj.tags = value.tags
        metadata_obj.description = value.description
        metadata_obj.read_permissions = cast(
            bytes,
            syft.serialize(syft.lib.python.Dict(value.read_permissions), to_bytes=True),
        ).hex()
        metadata_obj.search_permissions = cast(
            bytes,
            syft.serialize(
                syft.lib.python.Dict(value.search_permissions), to_bytes=True
            ),
        ).hex()
        metadata_obj.write_permissions = cast(
            bytes,
            syft.serialize(
                syft.lib.python.Dict(value.write_permissions), to_bytes=True
            ),
        ).hex()

        obj_dataset_relation = self._get_obj_dataset_relation(key)
        if obj_dataset_relation:
            # Create a object dataset relationship for the new object
            obj_dataset_relation = BinObjDataset(
                # id=obj_dataset_relation.id,  NOTE: Commented temporarily
                name=obj_dataset_relation.name,
                obj=str(key.value),
                dataset=obj_dataset_relation.dataset,
                dtype=obj_dataset_relation.dtype,
                shape=obj_dataset_relation.shape,
            )

        local_session = sessionmaker(bind=self.db)()
        if create_metadata:
            local_session.add(metadata_obj)
        local_session.add(obj_dataset_relation) if obj_dataset_relation else None
        local_session.commit()
        local_session.close()

    def delete(self, key: UID) -> None:
        try:
            self.redis.delete(str(key.value))
            local_session = sessionmaker(bind=self.db)()
            metadata_to_delete = (
                local_session.query(ObjectMetadata)
                .filter_by(obj=str(key.value))
                .first()
            )
            local_session.delete(metadata_to_delete)
            local_session.commit()
            local_session.close()
        except Exception as e:
            print(f"{type(self)} Exception in __delitem__ error {key}. {e}")

    def clear(self) -> None:
        self.redis.flushdb()
        local_session = sessionmaker(bind=self.db)()
        local_session.query(ObjectMetadata).delete()
        local_session.commit()
        local_session.close()

    def __repr__(self) -> str:
        return f"{type(self)}"
